// exec.hh
// executor class
// Kevin Puckett Albrecht
// Rev 22
// 10 April 2000 - 6 May 2001
//
// must define: GLOBAL_FILE_LINE_SIZE

#ifndef EXEC_HH
#define EXEC_HH

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdlib.h>
#include <time.h>

// Local //
#include "tree.hh"
#include "icode.hh"
#include "token.hh"
#include "common.hh"
#include "runstack.hh"
#include "filetree.hh"

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

class Exec
{
  public:
    Exec (int anMaxGotos)
    {
      mpFileTree = NULL;
      mnStatements = 0;
      mnMaxGotos = anMaxGotos;
    }
    void Execute ();

  private:
    FileTree *mpFileTree; // holds the files (not allocated unless needed)
    Token *mpToken;       // current token
    TokenCode mCode;      // code of current token
    TreeNode *mpNode;     // tree node
    long mnStatements;    // number of executed statements
    RunStack mStack;
    int mnMaxGotos;       // maximum number of gotos permitted

    void GetToken ()
    {
      mpToken = gpICode->Get ();
      mCode = mpToken->code;
      mpNode = gpICode->GetNode ();
    }
    void GoTo (int anPos)
    {
      gpICode->GoTo (anPos);
    }
    int CurrentPos () const
    {
      return gpICode->GetCurrentPos ();
    }

    // Structures //
    void ExecStatementList (TokenCode codeListToken = codeDummy, TokenCode codeEndingToken = codeEND);

    // Statements //
    void ExecStatement ();
    void ExecAssignment ();

    void ExecLabel ();
    void ExecDeclareVar ();
    void ExecBlock ();
    void ExecGoto ();
    void ExecIf ();
    void ExecLoop ();
    void ExecDeclarative ();
    void ExecQuit ();

    // Console I/O //
    void ExecPut ();
    void ExecGetline ();

    // Math //
    void ExecMrand ();

    // String Manipulation //
    void ExecSnext ();
    void ExecSzero ();
    void ExecSget ();

    // File I/O //
    void ExecFin ();
    void ExecFout ();
    void ExecFgetline ();
    void ExecFput ();
    void ExecFclose ();

    // Expressions //
    void ExecExpression ();
    void ExecSimpleExpression ();
    void ExecTerm ();
    void ExecFactor ();
};

//// Execute ////
void Exec::Execute ()
{
  // Randomize timer for "mrand" //
  srand ((unsigned) time (NULL));
  srand (((unsigned) time (NULL) * (rand () % 90)) + 1);

  // Reset the icode to the start //
  gpICode->Reset ();

  // The parser enters the start label with a marker of '-1';  if //
  //  it is still '-1', start at the beginning of the program;    //
  //  otherwise, goto that position.                              //
  if (gpStartNode->marker != -1)
  {
    GoTo (gpStartNode->marker);
  }

  GetToken (); // now first token of program

  do
  {
    ExecStatement ();

    // Skip any blank lines //
    while (mCode == codeEOL) GetToken ();

  } while (mCode != codeEOF);
}

//// ExecStatementList ////
void Exec::ExecStatementList (TokenCode codeListToken, TokenCode codeEndingToken)
{
  do
  {
    ExecStatement ();

    // Skip any blank lines //
    while (mCode == codeEOL) GetToken ();

    // `end` token reached; skip it and stop the statement list //
    if (mCode == codeEndingToken)
    {
      GetToken (); // now first token after `end`
      break;
    }
  } while (mCode != codeEOF);
}

//// ExecStatement ////
void Exec::ExecStatement ()
{
  switch (mCode)
  {
    case (codeIdentifier):
      ExecAssignment ();
      break;
    case (codeCOLON):
      ExecLabel ();
      break;
    case (codeVAR):
      ExecDeclareVar ();
      break;
    case (codeBLOCK):
      ExecBlock ();
      break;
    case (codeGOTO):
      ExecGoto ();
      break;
    case (codeIF):
      ExecIf ();
      break;
    case (codeLOOP):
      ExecLoop ();
      break;
    case (codeQUESTIONMARK):
      ExecDeclarative ();
      break;
    case (codeQUIT):
      ExecQuit ();
      break;

    case (codePUT):
    case (codePUTN):
      ExecPut ();
      break;
    case (codeGETLINE):
      ExecGetline ();
      break;

    case (codeMRAND):
      ExecMrand ();
      break;

    case (codeSNEXT):
      ExecSnext ();
      break;
    case (codeSZERO):
      ExecSzero ();
      break;
    case (codeSGET):
      ExecSget ();
      break;

    case (codeFIN):
      ExecFin ();
      break;
    case (codeFOUT):
      ExecFout ();
      break;
    case (codeFGETLINE):
      ExecFgetline ();
      break;
    case (codeFPUT):
    case (codeFPUTN):
      ExecFput ();
      break;
    case (codeFCLOSE):
      ExecFclose ();
      break;
    case (codeEND):
      gpErrorR->Show (EFF_TERMINATE, ER_INVALID_EXEC_POINT,
                      gvsTokenStrings[mCode], "Goto statement jumped into a nested scope");
      break;
    case (codeEOL):
    case (codeEOF):
      break;
    default:
      gpErrorR->Show (EFF_TERMINATE, ER_INVALID_EXEC_POINT,
                      gvsTokenStrings[mCode]);
      break;
  }
}

//// ExecAssignment ////
void Exec::ExecAssignment ()
{
  TreeNode *pNode = mpNode;

  // Through the use of 'goto' to jump to a label or section,
  //  it is posible to bypass the original definition of a
  //  variable.  Note that this should be ignored if variable
  //  declarations are implicit.
  if ((mpNode->bDefined == false) && (gpCommon->bImplicit == false))
  {
    gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                    "Execution jumped to a point after the definition"
                   );
  }

  GetToken (); // now `:` or `@`

  if (mCode == codeANDSIGN)
  {
    int nPos = 0;
    GetToken (); // now first token of position expression

    // Execute position expression and pop the value //
    ExecExpression ();
    switch (pNode->type)
    {
      case (tValue):
        // One is taken away because Twist counting //
        //  ... begins at one, not zero, as in C++. //
        nPos = Value::IntD (mStack.Pop()) - 1;
        break;
      default:
        gpErrorR->Show (EFF_TERMINATE, ER_INVALID_ASSIGNMENT, NULL,
            k_app ("Type not recognized in assignments: ", gvsTypeStrings[pNode->type])
            );
        break;
    }

    ; // now `:`
    GetToken (); // now first token of assignment expression

    // Execute assignment expression and pop the value //
    ExecExpression ();

    // Set the char in nPos to the first char in expression //
    char *sExpression = mStack.Pop();
    char *sValue = pNode->GetValue();
    sValue[nPos] = sExpression[0];
    pNode->SetValueD (sValue);
    delete [] sExpression;
  }
  else
  {
    GetToken (); // now first token of expression

    // Execute assignment expression and pop the value //
    ExecExpression ();
    switch (pNode->type)
    {
      case (tValue):
        pNode->SetValueD (mStack.Pop());
        break;
      default:
        gpErrorR->Show (EFF_TERMINATE, ER_INVALID_ASSIGNMENT, NULL,
            k_app ("Type not recognized in assignments: ", gvsTypeStrings[pNode->type])
            );
        break;
    }
  }
}

//// ExecLabel ////
void Exec::ExecLabel ()
{
  // NOTE: This procedure really does nothing. All data is entered //
  //       into the node upon parsing.                             //

  GetToken (); // now identifier
  GetToken (); // now EOL or first token of statement

  if (mCode != codeEOL)
  {
    ExecStatement ();
  }
}

//// ExecDeclareVar ////
void Exec::ExecDeclareVar ()
{
  GetToken (); // now identifier

  TreeNode *pNode = mpNode; // must capture the node when identifier

  if (pNode->bDefined == true)
  {
    /***
    ** gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_ALREADY_DEFINED, mpToken->buffer,
    **                "Execution jumped to a point before the definition after passing it"
    **               );
    ***/

    // Whenever the execution of the program passes a variable  //
    //  declararation, the value in the node should be reset to //
    //  empty.  That is what the next line does.                //
    pNode->SetValue ("");
  }
  else
  {
    pNode->bDefined = true;
  }

  GetToken (); // now `:` or EOL
  if (mCode == codeEOL) return;

  GetToken (); // now first token of expression

  // Execute expression and pop the value //
  ExecExpression ();
  switch (pNode->type)
  {
    case (tValue):
      pNode->SetValueD (mStack.Pop());
      break;
    default:
      gpErrorR->Show (EFF_TERMINATE, ER_INVALID_ASSIGNMENT, NULL,
                      k_app ("Type not recognized in assignments: ", gvsTypeStrings[pNode->type])
                     );
      break;
  }
}

//// ExecBlock ////
void Exec::ExecBlock ()
{
  GetToken (); // now EOL

  ExecStatementList ();
}

//// ExecGoto ////
void Exec::ExecGoto ()
{
  ++gpCommon->nGotos;

  // Cause error if too many gotos //
  if (gpCommon->nGotos >= mnMaxGotos)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_EXCESSIVE_GOTOS, k_int (gpCommon->nGotos));
  }

  GetToken (); // now identifier

  // Goto the position of the label //
  GoTo (mpNode->marker);

  GetToken (); // now EOL
}

//// ExecIf ////
void Exec::ExecIf ()
{
  GetToken (); // now marker token
  int nEndPos = mpToken->marker;

  GetToken (); // now first token of boolean expression

  // Evaluate boolean expression //
  ExecExpression ();

  if (Value::BoolD (mStack.Pop()) == false)
  {
    gpICode->GoTo (nEndPos);
    GetToken (); // now first token after `end`
  }
  else
  {
    ExecStatementList ();
  }
}

//// ExecLoop ////
void Exec::ExecLoop ()
{
  GetToken (); // now marker token
  int nEndPos = mpToken->marker;
  int nStartPos = CurrentPos ();

  GetToken (); // now first token of boolean expression

  while (true)
  {
    ExecExpression ();

    if (Value::BoolD (mStack.Pop ()) == false)
    {
      gpICode->GoTo (nEndPos);
      GetToken (); // now first token after `end`
      break;
    }

    ExecStatementList ();
    gpICode->GoTo (nStartPos);
    GetToken (); // now first token of boolean expression
  }
}

//// ExecDeclarative ////
void Exec::ExecDeclarative ()
{
  // This section does nothing; all declarative actions occur //
  //  in the parser.                                          //

  GetToken (); // now first declarative token or EOL

  while (mCode != codeEOL)
  {
    GetToken (); // now next declarative token or EOL
  }
}

//// ExecQuit ////
void Exec::ExecQuit ()
{
  GetToken (); // now EOL
  exit (0);
}

//// ExecPut ////
void Exec::ExecPut ()
{
  TokenCode code = mCode;

  GetToken (); // now EOL or first token of expression

  if (mCode == codeEOL)
  {
    if (code == codePUTN) cout << endl;
  }
  else
  {
    // Evaluate and print the value of the expression //
    ExecExpression ();
    char *temp = mStack.Pop();
    cout << temp;
    k_del (temp);
    if (code == codePUTN) cout << endl;
  }
}

//// ExecGetline ////
void Exec::ExecGetline ()
{
  GetToken (); // now identifier

  // NOTE: This section uses strings for input to allow cin to //
  //  understand.  If float is not used, cin thinks that the   //
  //  '.' in '0.2' means that there are two integers being     //
  //  entered.                                                 //

  // Get input from user //
  char *sInput = new char[256];
  k_copy (sInput, "");
  cin.get (sInput, 256);
  cin.ignore (256, '\n');

  // Through the use of 'goto' to jump to a label or section,
  //  it is posible to bypass the original definition of a
  //  variable.  Note that this should be ignored if variable
  //  declarations are implicit.
  if ((mpNode->bDefined == false) && (gpCommon->bImplicit == false))
  {
    gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                    "Execution jumped to a point after the definition"
                   );
  }

  // Put input in the node //
  mpNode->SetValueD (sInput);

  GetToken (); // now EOL
}

//// ExecMrand ////
void Exec::ExecMrand ()
{
  GetToken (); // now identifier
  TreeNode *pNode = mpNode;

  // Through the use of 'goto' to jump to a label or section,
  //  it is posible to bypass the original definition of a
  //  variable.  Note that this should be ignored if variable
  //  declarations are implicit.
  if ((mpNode->bDefined == false) && (gpCommon->bImplicit == false))
  {
    gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                    "Execution jumped to a point after the definition"
                   );
  }

  GetToken (); // now comma
  GetToken (); // now first token of lower bound expression

  // Get lower bound expression //
  ExecExpression ();
  int nLower = Value::IntD (mStack.Pop());

  GetToken (); // now first token of upper bound expression

  // Get upper bound expression //
  ExecExpression ();
  int nUpper = Value::IntD (mStack.Pop());

  Error::Verbose (Value::String(nLower), Value::String(nUpper));

  int nAnswer = (rand () % (nUpper - nLower + 1)) + nLower;
  pNode->SetValue (Value::String (nAnswer));

  GetToken (); // now EOL
}

//// ExexSnext ////
void Exec::ExecSnext ()
{
  GetToken (); // now identifier
  TreeNode *pNode = mpNode;

  // Through the use of 'goto' to jump to a label or section,
  //  it is posible to bypass the original definition of a
  //  variable.
  if (pNode->bDefined == false)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                    "Execution jumped to a point after the definition"
                   );
  }

  GetToken (); // now comma
  GetToken (); // now first token of expression

  ExecExpression ();
  char *sSearch = mStack.Pop();
  char *sSource = pNode->GetValue();

  int nPosOfSearch = k_pos (sSource, sSearch, pNode->pos);
  if (nPosOfSearch == -1)
  {
    gpAutoNode->SetValue ("snext 0");
  }
  else
  {
    pNode->pos = nPosOfSearch + strlen (sSearch);
    gpAutoNode->SetValue ("");
  }

  // At this point, the position marker in the value is at the
  //  position after the search string.

  k_del (sSearch);
  k_del (sSource);

  GetToken (); // now EOL
}

//// ExecSzero ////
void Exec::ExecSzero ()
{
  GetToken (); // now identifier
  TreeNode *pNode = mpNode;

  // Through the use of 'goto' to jump to a label or section,
  //  it is posible to bypass the original definition of a
  //  variable.
  if (pNode->bDefined == false)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                    "Execution jumped to a point after the definition"
                   );
  }

  pNode->pos = 0;

  GetToken (); // now EOL
}

//// ExecSget ////
void Exec::ExecSget ()
{
  GetToken (); // now identifier
  TreeNode *pSourceNode = mpNode;

  // Through the use of 'goto' to jump to a label or section,
  //  it is posible to bypass the original definition of a
  //  variable.
  if (pSourceNode->bDefined == false)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                    "Execution jumped to a point after the definition"
                   );
  }

  GetToken (); // now comma

  GetToken (); // now identifier
  TreeNode *pDestNode = mpNode;

  // Through the use of 'goto' to jump to a label or section,
  //  it is posible to bypass the original definition of a
  //  variable.
  if (pDestNode->bDefined == false)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                    "Execution jumped to a point after the definition"
                   );
  }

  GetToken (); // now comma
  GetToken (); // now first token of expression

  ExecExpression ();
  char *sSearch = mStack.Pop();
  char *sSource = pSourceNode->GetValue();
  char *sResult;

  int nPosOfSearch = k_pos (sSource, sSearch, pSourceNode->pos);
  if (nPosOfSearch == -1)
  {
    sResult = "";
    gpAutoNode->SetValue ("get 0");
  }
  else
  {
    sResult = k_sub (sSource, pSourceNode->pos, nPosOfSearch);
    pSourceNode->pos = nPosOfSearch;
    gpAutoNode->SetValue ("");
  }

  // At this point, the position marker in the value is at the
  //  position before the the search string.

  pDestNode->SetValue (sResult);

  k_del (sSearch);
  k_del (sSource);
  k_del (sResult);

  GetToken (); // now EOL
}

//// ExecFin ////
void Exec::ExecFin ()
{
  // If mpFileTree not yet initialized, do so //
  if (!mpFileTree) mpFileTree = new FileTree;

  GetToken (); // now first token of expression

  // Get filename expression //
  ExecExpression ();
  char *sFilename = mStack.Pop();

  // Get file vertex //
  FileVertex *pTemp = mpFileTree->EnterNew (sFilename);
  if (pTemp)
  {
    pTemp->filehandle.open (sFilename, ios::in | ios::nocreate);
    if (!pTemp->filehandle.good())
    {
      gpAutoNode->SetValue ("fin 0");
    }
    else
    {
      pTemp->bOpen = true;
      pTemp->bInput = true;
      gpAutoNode->SetValue ("");
    }
  }
  else
  {
    gpAutoNode->SetValue ("fin opened-already");
  }

  k_del (sFilename);
}

//// ExecFout ////
void Exec::ExecFout ()
{
  // If mpFileTree not yet initialized, do so //
  if (!mpFileTree) mpFileTree = new FileTree;

  GetToken (); // now first token of expression

  // Get filename expression //
  ExecExpression ();
  char *sFilename = mStack.Pop();

  // Get file vertex //
  FileVertex *pTemp = mpFileTree->EnterNew (sFilename);
  if (pTemp)
  {
    pTemp->filehandle.open (sFilename, ios::out);
    if (!pTemp->filehandle.good())
    {
      gpAutoNode->SetValue ("fout 0");
    }
    else
    {
      pTemp->bOpen = true;
      pTemp->bInput = false; // output
      gpAutoNode->SetValue ("");
    }
  }
  else
  {
    gpAutoNode->SetValue ("fout opened-already");
  }

  k_del (sFilename);
}

//// ExecFgetline ////
void Exec::ExecFgetline ()
{
  // If mpFileTree not yet initialized, do so //
  if (!mpFileTree) mpFileTree = new FileTree;

  GetToken (); // now first token of filename expression

  // Get filename expression //
  ExecExpression ();
  char *sFilename = mStack.Pop();

  // Get file vertex //
  FileVertex *pTemp = mpFileTree->Search (sFilename);
  if (!pTemp)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_FILE_NOT_OPEN, sFilename);
  }

  // Verify i/o status of filehandle //
  if (pTemp->bInput != true)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_ATTEMPT_INPUT_ON_OUTFILE, sFilename);
  }

  ; // now comma
  GetToken (); // now identifier

  // Through the use of 'goto' to jump to a label or section,
  //  it is posible to bypass the original definition of a
  //  variable.  Note that this should be ignored if variable
  //  declarations are implicit.
  if ((mpNode->bDefined == false) && (gpCommon->bImplicit == false))
  {
    gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                    "Execution jumped to a point after the definition"
                   );
  }

  // Get input from file //
  char *sLine = new char[GLOBAL_FILE_LINE_SIZE];
  k_copy (sLine, "");

  // Get the line, and put it in the node //
  if (pTemp->filehandle.getline (sLine, GLOBAL_FILE_LINE_SIZE)) // not EOF yet
  {
    // Put input in the node //
    mpNode->SetValueD (sLine);
    gpAutoNode->SetValue ("");
  }
  else // EOF reached
  {
    gpAutoNode->SetValue ("fgetline eof");
  }

  GetToken (); // now EOL

  k_del (sFilename);
  k_del (sLine);
}

//// ExecFput ////
void Exec::ExecFput ()
{
  // If mpFileTree not yet initialized, do so //
  if (!mpFileTree) mpFileTree = new FileTree;

  TokenCode code = mCode;

  GetToken (); // now first token of filename expression

  // Get filename expression //
  ExecExpression ();
  char *sFilename = mStack.Pop();

  // Get file vertex //
  FileVertex *pTemp = mpFileTree->Search (sFilename);
  if (!pTemp)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_FILE_NOT_OPEN, sFilename);
  }

  // Verify i/o status of filehandle //
  if (pTemp->bInput != false)
  {
    gpErrorR->Show (EFF_TERMINATE, ER_ATTEMPT_OUTPUT_ON_INFILE, sFilename);
  }

  GetToken (); // now comma

  GetToken (); // now first token of output expression

  // Get output expression //
  ExecExpression ();
  char *sOutput = mStack.Pop();

  // Output the expression to file //
  pTemp->filehandle << sOutput;
  if (code == codeFPUTN) pTemp->filehandle << endl;

  GetToken (); // now EOL

  k_del (sFilename);
  k_del (sOutput);
}

//// ExecFclose ////
void Exec::ExecFclose ()
{
  // If mpFileTree not yet initialized, do so //
  if (!mpFileTree) mpFileTree = new FileTree;

  GetToken (); // now first token of expression

  // Get filename expression //
  ExecExpression ();
  char *sFilename = mStack.Pop();

  // Get file vertex //
  FileVertex *pTemp = mpFileTree->Search (sFilename);
  if (pTemp)
  {
    pTemp->filehandle.close();
    pTemp->bOpen = false;
    pTemp->bInput = true; // default to true
  }
  else
  {
    gpAutoNode->SetValue ("fclosed closed-already");
  }
}

//// ExecExpression ////
void Exec::ExecExpression ()
{
  TokenCode oper;

  ExecSimpleExpression ();

  // If there is a relational operator, execute another simple expression //
  if ((mCode == codeIS) || (mCode == codeISNOT) ||
      (mCode == codeEQUAL) || (mCode == codeNOTEQUAL) ||
      (mCode == codeLT) || (mCode == codeGT) ||
      (mCode == codeLTEQUAL) || (mCode == codeGTEQUAL) ||
      (mCode == codeANDSIGN))
  {
    oper = mCode;

    GetToken ();
    ExecSimpleExpression ();

    // Pop off the operands //
    char *sOperand2 = mStack.Pop();
    char *sOperand1 = mStack.Pop();

    float fOperand2 = Value::Float (sOperand2);
    float fOperand1 = Value::Float (sOperand1);

    int nOperand2 = Value::Int (sOperand2);

    // Push the result of the operation on the stack //
    switch (oper)
    {
      case (codeIS):
      {
        mStack.PushD (Value::String (k_same (sOperand1, sOperand2) ? true : false));
        break;
      }
      case (codeISNOT):
      {
        mStack.PushD (Value::String (k_same (sOperand1, sOperand2) ? false : true));
        break;
      }
      case (codeEQUAL):
      {
        mStack.PushD (Value::String ((fOperand1 == fOperand2) ? true : false));
        break;
      }
      case (codeNOTEQUAL):
      {
        mStack.PushD (Value::String ((fOperand1 != fOperand2) ? true : false));
        break;
      }
      case (codeLT):
      {
        mStack.PushD (Value::String ((fOperand1 < fOperand2) ? true : false));
        break;
      }
      case (codeGT):
      {
        mStack.PushD (Value::String ((fOperand1 > fOperand2) ? true : false));
        break;
      }
      case (codeLTEQUAL):
      {
        mStack.PushD (Value::String ((fOperand1 <= fOperand2) ? true : false));
        break;
      }
      case (codeGTEQUAL):
      {
        mStack.PushD (Value::String ((fOperand1 >= fOperand2) ? true : false));
        break;
      }
      case (codeANDSIGN):
      {
        mStack.PushD (Value::String (sOperand1[nOperand2 - 1]));
        break;
      }
    }

    k_del (sOperand1);
    k_del (sOperand2);
  }
}

//// ExecSimpleExpression ////
void Exec::ExecSimpleExpression ()
{
  TokenCode operBi; // binary operator
  TokenCode operUn = codePLUS; // unary operator

  // Unary + or - //
  if ((mCode == codePLUS) || (mCode == codeMINUS))
  {
    operUn = mCode;
    GetToken ();
  }

  // Execute the first term and negate the value //
  ExecTerm ();
  if (operUn == codeMINUS)
  {
    mStack.Push (Value::String (-Value::FloatD (mStack.Pop ())));
  }
  
  // Loop to execute multiple additive operators //
  while ((mCode == codeMINUS) || (mCode == codePLUS) ||
         (mCode == codeAPPEND) || (mCode == codeOR)
        )
  {
    operBi = mCode;

    GetToken ();
    ExecTerm ();

    // Pop off the operands //
    char *sOperand2 = mStack.Pop();
    char *sOperand1 = mStack.Pop();

    float fOperand2 = Value::Float (sOperand2);
    float fOperand1 = Value::Float (sOperand1);
    bool bOperand2 = Value::Bool (sOperand2);
    bool bOperand1 = Value::Bool (sOperand1);

    // Push the result of the operation on the stack //
    switch (operBi)
    {
      case (codePLUS):
      {
        mStack.PushD (Value::String (fOperand1 + fOperand2));
        break;
      }
      case (codeMINUS):
      {
        mStack.PushD (Value::String (fOperand1 - fOperand2));
        break;
      }
      case (codeAPPEND):
      {
        char *sBuffer = new char[GLOBAL_VALUE_STR_SIZE];
        k_copy (sBuffer, sOperand1);
        k_cat (sBuffer, sOperand2);
        mStack.PushD (sBuffer);
        break;
      }
      case (codeOR):
      {
        mStack.PushD (Value::String ((bOperand1 || bOperand2) ? true : false));
        break;
      }
    }

    k_del (sOperand1);
    k_del (sOperand2);
  }    
}

//// ExecTerm ////
void Exec::ExecTerm ()
{
  TokenCode oper;

  // Execute the first factor //
  ExecFactor ();

  // Loop to execute multiple multiplicative operators //
  while ((mCode == codeTIMES) || (mCode == codeDIVIDE) || (mCode == codeAND))
  {
    oper = mCode;

    GetToken ();
    ExecFactor ();

    // Pop off the operands //
    char *sOperand2 = mStack.Pop();
    char *sOperand1 = mStack.Pop();

    float fOperand2 = Value::Float (sOperand2);
    float fOperand1 = Value::Float (sOperand1);
    bool bOperand2 = Value::Bool (sOperand2);
    bool bOperand1 = Value::Bool (sOperand1);

    // Push the result of the operation on the stack //
    switch (oper)
    {
      case (codeTIMES):
      {
        mStack.PushD (Value::String (fOperand1 * fOperand2));
        break;
      }
      case (codeDIVIDE):
      {
        if (fOperand2 != 0.0)
        {
          mStack.PushD (Value::String (fOperand1 / fOperand2));
        }
        else
        {
          gpErrorR->Show (EFF_TERMINATE, ER_DIVIDE_BY_ZERO);
        }
        break;
      }
      case (codeAND):
      {
        mStack.PushD (Value::String ((bOperand1 && bOperand2) ? true : false));
        break;
      }
    }

    k_del (sOperand1);
    k_del (sOperand2);
  }    
}

//// ExecFactor ////
void Exec::ExecFactor ()
{
  switch (mCode)
  {
    case (codeIdentifier):
    {
      // Through the use of 'goto' to jump to a label or section,
      //  it is posible to bypass the original definition of a
      //  variable.  Note that this should be ignored if variable
      //  declarations are implicit.
      if ((mpNode->bDefined == false) && (gpCommon->bImplicit == false))
      {
        gpErrorR->Show (EFF_TERMINATE, ER_VARIABLE_NOT_DEFINED, mpToken->buffer,
                        "Execution jumped to a point after the definition"
                       );
      }

      // Push the variable's value onto the stack //
      mStack.Push (mpNode->GetValue ());

      GetToken ();
      break;
    }
    case (codeNumber):
    {
      // Push the number's value onto the runtime stack //
      mStack.Push (mpNode->GetValue ());

      GetToken ();
      break;
    }
    case (codeString):
    {
      // Push the string's value onto the runtime stack //
      mStack.Push (mpNode->GetValue ());

      GetToken ();
      break;
    }
    case (codeNOT):
    {
      // execute factor and invert it's value //
      GetToken ();
      ExecFactor ();

      mStack.PushD (Value::String (Value::BoolD (mStack.Pop()) ? false : true));
      break;
    }
    case (codeINT):
    {
      // execute factor and convert it's value //
      GetToken ();
      ExecFactor ();

      mStack.PushD (Value::String (Value::IntD (mStack.Pop())));
      break;
    }
    case (codeRAT):
    {
      // execute factor and convert it's value //
      GetToken ();
      ExecFactor ();

      mStack.PushD (Value::String (Value::FloatD (mStack.Pop())));
      break;
    }
    case (codeSTR):
    {
      //NOTE: this section is here only for completeness //

      // execute factor and convert it's value //
      GetToken ();
      ExecFactor ();

      mStack.PushD (mStack.Pop());
      break;
    }
    case (codeBOOL):
    {
      // execute factor and convert it's value //
      GetToken ();
      ExecFactor ();

      mStack.PushD (Value::String (Value::BoolD (mStack.Pop())));
      break;
    }
    case (codeLEFTPAREN):
    {
      GetToken (); // now first token after `(`
      ExecExpression ();
      GetToken (); // now first token after `)`
    }
  }
}

#endif