// parser.hh
// parser class
// Kevin Puckett Albrecht
// Rev 22
// 23 March 2000 - 6 May 2001

#ifndef PARSER_HH
#define PARSER_HH

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "common.hh"
#include "error.hh"
#include "token.hh"
#include "scanner.hh"
#include "runstack.hh"
#include "tree.hh"
#include "value.hh"
#include "treestack.hh"

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

class Parser
{
  public:
    Parser ()
    {
      mpScanner = new Scanner ();
      mbDeclarativeOver = false;
    }
    ~Parser ()
    {
      delete mpScanner;
    }

    void Parse ();

  private:
    Scanner *mpScanner;
    Token *mpToken; // current token
    TokenCode mCode; // code of current token
    TreeStack mStack; // the symbol table stack
    bool mbDeclarativeOver;

    // Parse Structures //
    void ParseStatementList (TokenCode codeEndingToken = codeEND);

    // Parse Statements //
    void ParseStatement ();
    void ParseAssignment ();
    void ParseLabel ();
    void ParseDeclareVar ();
    void ParseBlock ();
    void ParseGoto ();
    void ParseIf ();
    void ParseLoop ();
    void ParseDeclarative ();
    void ParseQuit ();

    // Console Output //
    void ParsePut ();
    void ParseGetline ();

    // Math //
    void ParseMrand ();

    // Parse String Manipulation //
    void ParseSnext ();
    void ParseSzero ();
    void ParseSget ();

    // Parse File I/O //
    void ParseFin ();
    void ParseFout ();
    void ParseFgetline ();
    void ParseFput ();
    void ParseFclose ();

    // Parse Expressions //
    void ParseExpression ();       // =, <>, <, >, <=, >=
    void ParseSimpleExpression (); // +u, -u, +, -, &, or
    void ParseTerm ();             // *, /, and
    void ParseFactor ();

    // Special Methods //
    void EnsureSkipEOL ();
    bool Sync (TokenCode aCode, char *asError = ES_UNEXPECTED_TOKEN);
    void SeekToEOL ();
    void GetToken ()
    {
      mpToken = mpScanner->Get ();
      mCode = mpToken->code;
    }
    void AppendToken ()
    {
      gpICode->Put (mCode);
    }
    void AppendLineMarker ()
    {
      gpICode->PutLineMarker ();
    }
    TreeNode *Find (char *asSearch)
    {
      return mStack.Find (asSearch);
    }
    TreeNode *SearchAll (char *asSearch)
    {
      return mStack.SearchAll (asSearch);
    }
    TreeNode *SearchLocal (char *asSearch)
    {
      return mStack.SearchLocal (asSearch);
    }
    TreeNode *EnterGlobal (char *asSearch, Type aType = tDummy)
    {
      return mStack.EnterGlobal (asSearch, aType);
    }
    TreeNode *EnterLocal (char *asSearch, Type aType = tDummy)
    {
      return mStack.EnterLocal (asSearch, aType);
    }
    TreeNode *EnterNewLocal (char *asSearch, Type aType = tDummy)
    {
      return mStack.EnterNewLocal (asSearch, aType);
    }
};

//// Parse ////
//// Parse the code
///////////////////
void Parser::Parse ()
{
  // Enter the start label with a marker of '-1';  if the executor //
  //  encounters a '-1' when it starts, it will start at the       //
  //  beginning of the program.                                    //
  gpStartNode = EnterGlobal ("start", tMarker);
  gpStartNode->SetContent (tMarker, false/*const*/, "", -1);

  //////////////////////
  // GLOBAL CONSTANTS //
  //////////////////////

  // Enter the bool values of true and false //
  TreeNode *pTrue = EnterGlobal ("true", tValue);
  pTrue->SetContent (tValue, true/*const*/, "1", 0);
  pTrue->bDefined = true;

  TreeNode *pFalse = EnterGlobal ("false", tValue);
  pFalse->SetContent (tValue, true/*const*/, "0", 0);
  pFalse->bDefined = true;

  // Enter the string value of break //
  TreeNode *pNL = EnterGlobal ("nl", tValue);
  pNL->SetContent (tValue, true/*const*/, "\n", 0);
  pNL->bDefined = true;

  // Enter the predefined value of the auto variable //
  gpAutoNode = EnterGlobal ("auto", tValue);
  gpAutoNode->SetContent (tValue, false/*constant*/, "", 0);
  gpAutoNode->bDefined = true;

  /////////////
  // PARSING //
  /////////////

  GetToken();
  AppendToken ();

  do {
    // Parse a Statement //
    ParseStatement ();

    // Handles empty files //
    if (mCode == codeEOF) break;

    EnsureSkipEOL();

  } while (mCode != codeEOF);
}

///////////////////////
// Structure Parsing //
///////////////////////
///////////////////////
///////////////////////

//// ParseStatementList ////
void Parser::ParseStatementList (TokenCode codeEndingToken)
{
  // Ending token reached; skip it and stop the statement list; //
  //  this conditional allows for empty statement lists         //
  if (mCode == codeEndingToken)
  {
    GetToken ();
    AppendToken ();
    return;
  }

  mStack.EnterScope();

  do {
    // Parse a Statement //
    ParseStatement ();

    EnsureSkipEOL ();

    // Ending token reached; skip it and stop the statement list //
    if (mCode == codeEndingToken)
    {
      mStack.ExitScope();

      GetToken ();
      AppendToken ();
      break;
    }
  } while (mCode != codeEOF);
}


///////////////////////
// Statement Parsing //
///////////////////////
///////////////////////
///////////////////////

//// ParseStatement ////
void Parser::ParseStatement ()
{
  AppendLineMarker ();

  switch (mCode)
  {
    case (codeIdentifier):
      ParseAssignment ();
      break;
    case (codeCOLON):
      ParseLabel ();
      break;
    case (codeVAR):
      ParseDeclareVar ();
      break;
    case (codeBLOCK):
      ParseBlock ();
      break;
    case (codeGOTO):
      ParseGoto ();
      break;
    case (codeIF):
      ParseIf ();
      break;
    case (codeLOOP):
      ParseLoop ();
      break;
    case (codeQUESTIONMARK):
      ParseDeclarative ();
      break;
    case (codeQUIT):
      ParseQuit ();
      break;

    case (codePUT):
    case (codePUTN):
      ParsePut ();
      break;
    case (codeGETLINE):
      ParseGetline ();
      break;

    case (codeMRAND):
      ParseMrand ();
      break;

    case (codeSNEXT):
      ParseSnext ();
      break;
    case (codeSZERO):
      ParseSzero ();
      break;
    case (codeSGET):
      ParseSget ();
      break;

    case (codeFIN):
      ParseFin ();
      break;
    case (codeFOUT):
      ParseFout ();
      break;
    case (codeFGETLINE):
      ParseFgetline ();
      break;
    case (codeFPUT):
    case (codeFPUTN):
      ParseFput ();
      break;
    case (codeFCLOSE):
      ParseFclose ();
      break;
    case (codeEOL):
    case (codeEOF):
      break;
    default:
      gpErrorS->Show (EFF_TERMINATE, ES_INVALID_PARSE_POINT,
                      gvsTokenStrings[mCode]);
      break;
  }
}

//// ParseAssignment ////
void Parser::ParseAssignment ()
{
  // Search for the target variable's identifier and enter it if necessary //
  TreeNode *pTargetNode = SearchAll (mpToken->buffer);
  if (!pTargetNode)
  {
    pTargetNode = EnterLocal (mpToken->buffer, tValue);

    if (gpCommon->bImplicit == false)
    {
      // Not found; Undefined identifier error; Enter it in symbol table //
      gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
    }
    else
    {
      // pNode->bInDeclaration is set to true for the next few lines //
      //  so that in ParseFactor, an error can be sounded if the     //
      //  same variable is used.  This prevents the value of the     //
      //  variable from a previous run of the variable declaration   //
      //  from influencing the current one.                          //
      pTargetNode->bInDeclaration = true;
      Error::Verbose ("PARS", "Variable implicitly declared", mpToken->buffer);
    }
  }
  else if (pTargetNode->bConst == true)
  {
    // Identifier is a constant; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_CONST_ASSIGNMENT, mpToken->buffer);
    SeekToEOL ();
    return;
  }
  gpICode->Put (pTargetNode);

  // : or @
  GetToken ();
  AppendToken ();

  // if ':', parse the expression //
  if (mCode == codeANDSIGN)
  {
    // Expression //
    GetToken ();
    AppendToken ();
    ParseExpression ();
  }

  // mCode must now be : //
  if (!Sync (codeCOLON, ES_MISSING_ASSIGNMENT_OP)) return;

  // Expression //
  GetToken ();
  AppendToken ();
  ParseExpression ();

  // In case implicit declaration is occuring, set the node //
  //  node off off of 'in declaration' mode                 //
  if ((gpCommon->bImplicit == true) && (pTargetNode->bInDeclaration == true))
  {
    pTargetNode->bInDeclaration = false;
  }
}

//// ParseLabel ////
void Parser::ParseLabel ()
{
  GetToken (); // now identifier
  AppendToken ();

  // Exit ParseLabel if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier; If not found, enter it; Append it //
  TreeNode *pNode = SearchAll (mpToken->buffer);
  if (pNode == NULL)
  {
    pNode = EnterGlobal (mpToken->buffer, tMarker);
  }
  else
  {
    if (pNode->type != tMarker)
    {
      gpErrorS->Show (EFF_RECOVER, ES_TYPE_INCOMPATIBLE, mpToken->buffer, "Identifier not defined as a label");
    }
  }
  gpICode->Put (pNode);

  // Set the node's marker with the current position //
  pNode->marker = gpICode->GetCurrentPos ();

  GetToken ();
  AppendToken ();

  if (mCode != codeEOL)
  {
    ParseStatement ();
  }
}

//// ParseDeclareVar ////
void Parser::ParseDeclareVar ()
{
  GetToken (); // now identifier
  AppendToken ();

  // Exit ParseDeclareVar if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Enter the identifier into the Tree; Append it //
  TreeNode *pNode = EnterNewLocal (mpToken->buffer, tValue);
  gpICode->Put (pNode);

  GetToken (); // `:` or EOL
  AppendToken ();

  if (mCode == codeEOL) // declaration only
  {
    return;
  }
  else // declaration and assignment
  {
    if (!Sync (codeCOLON, ES_MISSING_ASSIGNMENT_OP)) return;

    // pNode->bInDeclaration is set to true for the next few lines //
    //  so that in ParseFactor, an error can be sounded if the     //
    //  same variable is used.  This prevents the value of the     //
    //  variable from a previous run of the variable declaration   //
    //  from influencing the current one.                          //
    pNode->bInDeclaration = true;
    GetToken (); // now expression
    AppendToken ();
    ParseExpression ();
    pNode->bInDeclaration = false;
  }
}

//// ParseBlock ////
void Parser::ParseBlock ()
{
  GetToken (); // now EOL
  AppendToken ();

  EnsureSkipEOL ();

  // Statements //
  ParseStatementList ();
}

//// ParseGoto ////
void Parser::ParseGoto ()
{
  GetToken (); // now identifier
  AppendToken ();

  // Exit if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier //
  TreeNode *pNode = SearchAll (mpToken->buffer);
  if (pNode)
  {
    // Found, check if correct type //
    if (pNode->type != tMarker)
    {
      gpErrorS->Show (EFF_RECOVER, ES_TYPE_INCOMPATIBLE, mpToken->buffer, "Identifier not defined as a label");
    }
  }
  else
  {
    // Not found; Enter it in symbol table //
    pNode = EnterGlobal (mpToken->buffer, tMarker);
  }
  gpICode->Put (pNode);

  GetToken (); // now EOL
  AppendToken ();
}

//// ParseIf ////
void Parser::ParseIf ()
{
  // Append a placeholder to hold the location //
  // of the end of the statement list.         //
  int nPosOfEndPosMarker = gpICode->GetCurrentPos ();
  gpICode->PutNumber (0);

  GetToken (); // now first token in boolean expression
  AppendToken ();

  // Expression //
  ParseExpression ();

  EnsureSkipEOL ();

  // Statements //
  ParseStatementList ();

  // Append (at the placeholder) the true position //
  // of the end of the statement list.             //
  int nPosOfEndOfList = gpICode->GetCurrentPos ();
  gpICode->GoTo (nPosOfEndPosMarker);
  gpICode->PutNumber (nPosOfEndOfList);
  gpICode->GoTo (nPosOfEndOfList);
}

//// ParseLoop ////
void Parser::ParseLoop ()
{
  // Append a placeholder to hold the location //
  // of the end of the statement list.         //
  int nPosOfEndPosMarker = gpICode->GetCurrentPos ();
  gpICode->PutNumber (0);

  GetToken (); // now first token in boolean expression
  AppendToken ();

  // Expression //
  ParseExpression ();

  EnsureSkipEOL ();

  // Statements //
  ParseStatementList ();

  // Append (at the placeholder) the true position //
  // of the end of the statement list.             //
  int nPosOfEndOfList = gpICode->GetCurrentPos ();
  gpICode->GoTo (nPosOfEndPosMarker);
  gpICode->PutNumber (nPosOfEndOfList);
  gpICode->GoTo (nPosOfEndOfList);
}

//// ParseDeclarative ////
void Parser::ParseDeclarative ()
{
  if (mbDeclarativeOver == true)
  {
    gpErrorS->Show (EFF_RECOVER, ES_REPEAT_DECLARATIVE);
    GetToken (); // now first declarative token or EOL
    AppendToken ();
    return;
  }
  mbDeclarativeOver = true;

  GetToken (); // now first declarative token or EOL
  AppendToken ();

  while (mCode != codeEOL)
  {
    switch (mCode)
    {
      case (codeIMPLICIT):
        gpCommon->bImplicit = true;
        break;
      case (codeEXPLICIT):
        gpCommon->bImplicit = false;
        break;
      default:
        gpErrorS->Show (EFF_RECOVER, ES_UNEXPECTED_TOKEN, mpToken->buffer);
        break;
    }
    GetToken (); // now next declarative token or EOL
    AppendToken ();
  }
}

//// ParseQuit ////
void Parser::ParseQuit ()
{
  GetToken (); // now EOL
  AppendToken ();
}

//// ParsePut ////
//// Parses the 'put' or 'putn' commands.
/////////////////////////////////////////
void Parser::ParsePut ()
{
  GetToken (); // now EOL or first token of exp
  AppendToken ();

  // If EOL, return //
  if (mCode == codeEOL) return;

  // Otherwise, parse the expression //
  ParseExpression ();
}

//// ParseGetline ////
void Parser::ParseGetline ()
{
  GetToken (); // now identifier
  AppendToken ();

  // Exit if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier //
  TreeNode *pNode = SearchAll (mpToken->buffer);
  if (!pNode)
  {
    if (gpCommon->bImplicit == false)
    {
      // Not found; Undefined identifier error; Enter it in symbol table //
      gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
    }
    else Error::Verbose ("PARS", "Variable implicitly declared", mpToken->buffer);
    pNode = EnterLocal (mpToken->buffer, tValue);
  }
  else if (pNode->bConst == true)
  {
    // Identifier is a constant; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_CONST_ASSIGNMENT, mpToken->buffer);
    SeekToEOL ();
    return;
  }
  gpICode->Put (pNode);

  GetToken (); // now EOL
  AppendToken ();
}

//// ParseMrand ////
void Parser::ParseMrand ()
{
  GetToken (); // now identifier
  AppendToken ();

  // Exit if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier //
  TreeNode *pNode = SearchAll (mpToken->buffer);
  if (!pNode)
  {
    if (gpCommon->bImplicit == false)
    {
      // Not found; Undefined identifier error; Enter it in symbol table //
      gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
    }
    else Error::Verbose ("PARS", "Variable implicitly declared", mpToken->buffer);
    pNode = EnterLocal (mpToken->buffer, tValue);
  }

  if (pNode->type != tValue)
  {
    // Identifier is not of type tValue; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_TYPE_INCOMPATIBLE, mpToken->buffer);
    SeekToEOL ();
    return;
  }

  if (pNode->bConst == true)
  {
    // Identifier is a constant; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_CONST_ASSIGNMENT, mpToken->buffer);
    SeekToEOL ();
    return;
  }

  gpICode->Put (pNode);

  GetToken (); // now comma
  AppendToken ();
  if (!Sync (codeCOMMA)) return;

  GetToken (); // now first token of lower bound expression
  AppendToken ();
  ParseExpression ();

  ; // now comma
  if (!Sync (codeCOMMA)) return;

  GetToken (); // now first token of upper bound expression
  AppendToken ();
  ParseExpression ();

  ; // now EOL
}

//// ParseSnext ////
void Parser::ParseSnext ()
{
  GetToken (); // now identifier
  AppendToken ();

  // Exit if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier //
  TreeNode *pNode = SearchAll (mpToken->buffer);
  if (!pNode)
  {
    // Not found; Undefined identifier error; Enter it in symbol table //
    gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
    pNode = EnterLocal (mpToken->buffer, tValue);
  }
  else if (pNode->type != tValue)
  {
    // Identifier is not of type tValue; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_TYPE_INCOMPATIBLE, mpToken->buffer);
    SeekToEOL ();
    return;
  }
  gpICode->Put (pNode);

  GetToken (); // now comma
  AppendToken ();
  if (!Sync (codeCOMMA)) return;

  GetToken (); // now first token of expression
  AppendToken ();
  ParseExpression ();
}

//// ParseSzero ////
void Parser::ParseSzero ()
{
  GetToken (); // now identifier
  AppendToken ();

  // Exit if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier //
  TreeNode *pNode = SearchAll (mpToken->buffer);
  if (!pNode)
  {
    // Not found; Undefined identifier error; Enter it in symbol table //
    gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
    pNode = EnterLocal (mpToken->buffer, tValue);
  }
  else if (pNode->type != tValue)
  {
    // Identifier is not of type tValue; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_TYPE_INCOMPATIBLE, mpToken->buffer);
    SeekToEOL ();
    return;
  }
  gpICode->Put (pNode);

  GetToken (); // now EOL
  AppendToken ();
}

//// ParseSget ////
void Parser::ParseSget ()
{
  GetToken (); // now identifier
  AppendToken ();

  // Exit if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier //
  TreeNode *pSourceNode = SearchAll (mpToken->buffer);
  if (!pSourceNode)
  {
    // Not found; Undefined identifier error; Enter it in symbol table //
    gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
    pSourceNode = EnterLocal (mpToken->buffer, tValue);
  }
  else if (pSourceNode->type != tValue)
  {
    // Identifier is not of type tValue; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_TYPE_INCOMPATIBLE, mpToken->buffer);
    SeekToEOL ();
    return;
  }
  gpICode->Put (pSourceNode);

  GetToken (); // now comma
  AppendToken ();
  if (!Sync (codeCOMMA)) return;

  GetToken (); // now identifier
  AppendToken ();

  // Exit if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier //
  TreeNode *pDestNode = SearchAll (mpToken->buffer);
  if (!pDestNode)
  {
    // Not found; Undefined identifier error; Enter it in symbol table //
    gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
    pDestNode = EnterLocal (mpToken->buffer, tValue);
  }
  else if (pDestNode->type != tValue)
  {
    // Identifier is not of type tValue; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_TYPE_INCOMPATIBLE, mpToken->buffer);
    SeekToEOL ();
    return;
  }
  gpICode->Put (pDestNode);

  GetToken (); // now comma
  AppendToken ();
  if (!Sync (codeCOMMA)) return;

  GetToken (); // now first token of expression
  AppendToken ();
  ParseExpression ();
}

//// ParseFin ////
void Parser::ParseFin ()
{
  GetToken (); // first token of exp
  AppendToken ();

  // Parse the expression //
  ParseExpression ();
}

//// ParseFout ////
void Parser::ParseFout ()
{
  GetToken (); // first token of exp
  AppendToken ();

  // Parse the expression //
  ParseExpression ();
}

//// ParseFgetline ////
void Parser::ParseFgetline ()
{
  GetToken (); // now first token of filename expression
  AppendToken ();
  ParseExpression ();

  ; // now comma
  if (!Sync (codeCOMMA)) return;

  GetToken (); // now identifier
  AppendToken ();

  // Exit if not an identifier //
  if (!Sync (codeIdentifier)) return;

  // Search for the identifier //
  TreeNode *pNode = SearchAll (mpToken->buffer);
  if (!pNode)
  {
    if (gpCommon->bImplicit == false)
    {
      // Not found; Undefined identifier error; Enter it in symbol table //
      gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
    }
    else Error::Verbose ("PARS", "Variable implicitly declared", mpToken->buffer);
    pNode = EnterLocal (mpToken->buffer, tValue);
  }

  if (pNode->type != tValue)
  {
    // Identifier is not of type tValue; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_TYPE_INCOMPATIBLE, mpToken->buffer);
    SeekToEOL ();
    return;
  }

  if (pNode->bConst == true)
  {
    // Identifier is a constant; Sync to EOL and return //
    gpErrorS->Show (EFF_RECOVER, ES_CONST_ASSIGNMENT, mpToken->buffer);
    SeekToEOL ();
    return;
  }

  gpICode->Put (pNode);

  GetToken (); // now EOL
  AppendToken ();
}

//// ParseFput ////
void Parser::ParseFput ()
{
  GetToken (); // now first token of filename expression
  AppendToken ();
  ParseExpression ();

  ; // now comma
  AppendToken ();
  if (!Sync (codeCOMMA)) return;

  GetToken (); // now first token of output expression
  AppendToken ();
  ParseExpression ();
}

//// ParseFclose ////
void Parser::ParseFclose ()
{
  GetToken (); // first token of filename expression
  AppendToken ();

  // Parse the expression //
  ParseExpression ();
}

////////////////////////
// Expression Parsing //
////////////////////////
////////////////////////
////////////////////////

//// ParseExpression ////
//// Duties: simple expression or two simple expressions
////////////////////////////////////////////////////////
void Parser::ParseExpression ()
{
  // Parse a simple expression //
  ParseSimpleExpression ();

  // If there is a relational operator, parse another simple expression //
  if ((mCode == codeIS) || (mCode == codeISNOT) ||
      (mCode == codeEQUAL) || (mCode == codeNOTEQUAL) ||
      (mCode == codeLT) || (mCode == codeGT) ||
      (mCode == codeLTEQUAL) || (mCode == codeGTEQUAL) ||
      (mCode == codeANDSIGN))
  {
    GetToken ();
    AppendToken ();
    ParseSimpleExpression ();
  }
}

//// ParseSimpleExpression ////
//// Duties: binary + or - or &, unary + or -, `or`
///////////////////////////////////////////////////
void Parser::ParseSimpleExpression ()
{
  // Unary + or - //
  if ((mCode == codeMINUS) || (mCode == codePLUS))
  {
    GetToken ();
    AppendToken ();
  }

  // Parse the first term //
  ParseTerm ();

  // Loop to parse subsequent additive operators and terms //
  while ((mCode == codeMINUS) || (mCode == codePLUS) ||
         (mCode == codeAPPEND) || (mCode == codeOR)
        )
  {
    GetToken ();
    AppendToken ();
    ParseTerm ();
  }
}

//// ParseTerm ////
//// Duties: binary * or /, `and`
/////////////////////////////////
void Parser::ParseTerm ()
{
  // Parse the first factor //
  ParseFactor();

  // Loop to parse subsequent multiplicative operators and factors //
  while ((mCode == codeTIMES) || (mCode == codeDIVIDE) || (mCode == codeAND))
  {
    GetToken ();
    AppendToken ();
    ParseFactor ();
  }
}

//// ParseFactor ////
//// Duties: identifier, number, parenthesized expression, `not`
////////////////////////////////////////////////////////////////
void Parser::ParseFactor ()
{
  switch (mCode)
  {
    case codeIdentifier:
    {
      // Search for the identifier //
      TreeNode *pNode = SearchAll (mpToken->buffer);
      if (!pNode)
      {
        // Not found; Undefined identifier error; Enter it in symbol table //
        gpErrorS->Show (EFF_RECOVER, ES_UNDEFINED_IDENTIFIER, mpToken->buffer);
        pNode = EnterLocal (mpToken->buffer, tValue);
      }
      gpICode->Put (pNode);

      // Check if the variable is currently being defined //
      if (pNode->bInDeclaration == true)
      {
        gpErrorS->Show (EFF_RECOVER, ES_VARIABLE_IN_DECLARATION, mpToken->buffer);
      }

      GetToken ();
      AppendToken ();
      break;
    }
    case codeNumber:
    {
      // Search for the number and enter it in the symbol table //
      TreeNode *pNode = SearchAll (mpToken->buffer);
      if (!pNode)
      {
        pNode = EnterLocal (mpToken->buffer, mpToken->type);
        pNode->SetValue (mpToken->value);
      }

      // Append handle to icode //
      gpICode->Put (pNode);

      GetToken ();
      AppendToken ();
      break;
    }
    case codeString:
    {
      // Search for the string and enter it in the symbol table //
      TreeNode *pNode = SearchAll (mpToken->buffer);
      if (!pNode)
      {
        pNode = EnterLocal (mpToken->buffer, mpToken->type);
        pNode->SetValue (mpToken->value);
      }

      // Append handle to icode //
      gpICode->Put (pNode);

      GetToken ();
      AppendToken ();
      break;
    }
    case codeNOT:
    {
      GetToken ();
      AppendToken ();
      ParseFactor ();
      break;
    }
    case codeINT:
    case codeRAT:
    case codeSTR:
    case codeBOOL:
    {
      GetToken ();
      AppendToken ();
      ParseFactor ();
      break;
    }
    case codeLEFTPAREN:
    {
      // Parenthesized subexpression; Call ParseExpression recursively //
      GetToken ();
      AppendToken ();
      ParseExpression ();

      // Check for the closing right parenthesis //
      if (mCode == codeRIGHTPAREN)
      {
        GetToken ();
        AppendToken ();
      }
      else
      {
        gpErrorS->Show (EFF_IGNORE, ES_MISSING_RIGHT_PAREN);
      }

      break;
    }
    default:
    {
      gpErrorS->Show (EFF_IGNORE, ES_INVALID_EXPRESSION);
      break;
    }
  }
}

/////////////////////
// Special Methods //
/////////////////////
/////////////////////
/////////////////////

void Parser::EnsureSkipEOL ()
{
  // Make sure that an end of line has been reached //
  Sync (codeEOL, ES_MISSING_EOL);

  // Skip over this EOL and any blank lines before parsing the next statement //
  while (mCode == codeEOL)
  {
    GetToken ();
    AppendToken ();
  }
}

bool Parser::Sync (TokenCode aCode, char *asError)
{
  // Make sure that one of the codes has been reached //
  if (mCode != aCode)
  {
    gpErrorS->Show (EFF_RECOVER, asError, mpToken->buffer);
    SeekToEOL ();
    return false; // code did not match
  }
  return true; // code did match
}

void Parser::SeekToEOL ()
{
  // To allow the parser to continue, skip tokens until an EOL is reached //
  while ((mCode != codeEOL) && (mCode != codeEOF))
  {
    GetToken ();
  }
}

#endif
